
from logging import getLogger

import numpy as np

from extra_data import by_id, SourceNameError
from exdf.data_reduction import ReductionMethod


class ManualRemoval(ReductionMethod):
    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'Manual removal',
            'Allows to manually remove sources, keys or trains.')

        group.add_argument(
            '--remove-sources',
            metavar='SRC_GLOB', action='store', type=str, nargs='*',
            help='remove one or more sources')

        group.add_argument(
            '--remove-keys',
            metavar='SRC_GLOB,KEY_GLOB', action='store', type=str, nargs='*',
            help='remove one or more keys from sources')

        group.add_argument(
            '--remove-trains',
            metavar='SLICE_EXPR', action='store', type=str, nargs='*',
            help='remove one or more trains')

    def __init__(self, data, args):
        if args.remove_sources:
            # Easiest would be to copy file and just remove the source.
            for source_glob in args.remove_sources:
                self.remove_sources(source_glob)

        if args.remove_keys:
            # Easiest would be to copy file and just remove the key.
            for sourcekey_str in args.remove_keys:
                source_glob, key_glob = sourcekey_str.split(',')
                self.remove_keys(source_glob, key_glob)

        if args.remove_trains:
            for slice_expr in args.remove_trains:
                # TODO: Convert slice_expr top slice
                self.remove_trains('*', by_id[args.remove_trains])


class PurgeEmptySources(ReductionMethod):
    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'Purge empty sources',
            'Allows to remove sources (or part of sources) entirely that are '
            'part of the input data, but do not contain data entry for any '
            'train. ')

        group.add_argument(
            '--purge-empty-sources',
            action='store_true',
            help='entirely remove any sources that do not contain data '
                 'for any train')

        group.add_argument(
            '--purge-empty-index-groups',
            action='store_true',
            help='entirely remove any index groups that do not contain data '
                 'for any train.')

    def __init__(self, data, args):
        if args.purge_empty_sources:
            for source in data.all_sources:
                if data[source].data_counts(False).sum() == 0:
                    self.remove_sources(source)

        if args.purge_empty_index_groups:
            for source in data.instrument_sources:
                sd = data[source]

                for grp in sd.index_groups:
                    if sd.data_counts(False, grp).sum() == 0:
                        self.remove_keys(source, f'{grp}.*')


class PpuTrainSequences(ReductionMethod):  # p3160, r250
    log = getLogger('exdf.data_reduction.builtins.PpuTrainSequences')

    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'PPU train sequences',
            'Allows to reduce data to those trains covered by PPU trigger '
            'sequences.')

        group.add_argument(
            '--ppu-select-trains-by',
            metavar='SRC', action='store', type=str,
            help='only select those trains covered by trigger sequences of '
                 'the given PPU device')

        group.add_argument(
            '--ppu-sliced-sources',
            metavar='SRC_GLOB', action='store', type=str, nargs='*',
            default=['HED_DET_AGIPD500K2G/DET/*CH0:xtdf'],
            help='sources sliced by PPU train sequences, HED-AGIPD by default')

        group.add_argument(
            '--ppu-train-offset',
            metavar='INT', action='store', type=int, default=0,
            help='offset between the train IDs in the PPU trigger sequences '
                 'and the actual trains picked, 0 by default')

    def __init__(self, data, args):
        if not args.ppu_select_trains_by:
            return

        try:
            seq_start = data[args.ppu_select_trains_by,
                             'trainTrigger.sequenceStart.value'].ndarray()
        except (SourceNameError, KeyError):
            self.log.error(f'PPU source {args.ppu_select_trains_by} or a '
                           f'required key not found in data')
            return

        # Only support single trains for now.
        train_ids = np.unique(seq_start)[1:] + args.ppu_train_offset

        self.log.info(f'Found {len(train_ids)} PPU sequences')

        for train_id in train_ids:
            for source_glob in args.ppu_sliced_sources:
                self.select_trains(source_glob, by_id[train_id:train_id+1])


class AgipdGain(ReductionMethod):
    log = getLogger('exdf.data_reduction.builtins.AgipdGain')

    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'AGIPD gain',
            'Allows to compress, replace or remove AGIPD gain information.')

        remove_group = group.add_mutually_exclusive_group()

        remove_group.add_argument(
            '--agipd-gain-remove-if-fixed',
            action='store_true',
            help='remove the chunks backing analog gain information from '
                 'AGIPD image data if operating in fixed gain')

        remove_group.add_argument(
            '--agipd-gain-remove-definitely',
            action='store_true',
            help='remove the chunks backing analog gain information from '
                 'AGIPD image data no matter the gain mode')

        group.add_argument(
            '--agipd-gain-sources', action='store', type=str,
            metavar='SRC_GLOB',
            help='AGIPD sources for which to remove analog gain chunks, '
                 'detected using extra_data.components if left blank')

        group.add_argument(
            '--agipd-gain-control', action='store', type=str, metavar='SRC',
            help='AGIPD control source to obtain gain information from, '
                 'inferred from gain sources if left blank')

    @staticmethod
    def _find_agipd(data):
        from extra_data.components import AGIPD1M, AGIPD500K

        try:
            domain = AGIPD1M._find_detector_name(data)
        except SourceNameError:
            pass
        else:
            return domain

        try:
            domain = AGIPD500K._find_detector_name(data)
        except SourceNameError:
            pass
        else:
            return domain

    def __init__(self, data, args):
        if not any([args.agipd_gain_remove_if_fixed,
                    args.agipd_gain_remove_definitely]):
            return

        if not args.agipd_gain_sources:
            domain = self._find_agipd(data)

            if domain is None:
                self.log.error('No AGIPD detected using '
                               'extra_data.components.AGIPD*, please pass '
                               'explicit gain sources')
                return

            self.log.info(f'No detector specified, detected {domain} '
                          f'automatically')
            agipd_sources = f'{domain}/DET/*CH0:xtdf'

        else:
            agipd_sources = args.agipd_gain_sources

        if args.agipd_gain_remove_if_fixed:
            try:
                if not args.agipd_gain_control:
                    domain = agipd_sources[:agipd_sources.index('/')]

                    if 'HED' in domain:
                        # Unfortunately HED uses a different domain for
                        # its AGIPD control devices, hardcode it here
                        # for convenience.
                        # TODO: This could also be solved via a CalCat
                        # query.
                        domain = domain.replace('_DET_', '_EXP_')

                    control_source = f'{domain}/MDL/FPGA_COMP'
                else:
                    control_source = args.agipd_gain_control

                gain_mode = data[control_source,
                                 'gainModeIndex'].as_single_value()
            except (SourceNameError, KeyError):
                self.log.error(f'No control source {control_source} or '
                               f'required keys found in data')
                return

            if gain_mode == 0:
                self.log.error('Data taken in adaptive mode, aborting')
                return

        self.rechunk_keys(agipd_sources, 'image.data', (-1, 1, None, None))
        self.subslice_keys(agipd_sources, 'image.data', np.s_[0, :, :])


class LpdMini(ReductionMethod):
    @staticmethod
    def arguments(ap):
        group = ap.add_argument_group(
            'LPD Mini'
            'Allows to manipulate LPD Mini data.')

        group.add_argument(
            '--lpd-mini-select-slots',
            action='store', metavar='POS', type=int, nargs='+',
            help='remove the chunks not covering data of the specified LPD '
                 'mini slots expressed by a list of 1-based positions')

        group.add_argument(
            '--lpd-mini-sources',
            action='store', type=str, metavar='SRC_GLOB',
            default='FXE_DET_LPD_MINI/DET/0CH0:xtdf',
            help='')

    def __init__(self, data, args):
        if not args.lpd_mini_select_slots:
            return

        lpdmini_sources = args.lpd_mini_sources

        self.rechunk_keys(lpdmini_sources, 'image.data', (-1, 32, None))

        for mini_pos in args.lpd_mini_select_slots:
            self.subslice_keys(lpdmini_sources, 'image.data',
                               np.s_[(mini_pos-1)*32:mini_pos*32, :, :])
